class TaskParseError(Exception):
    """
    Represents an error when parsing a task.
    """

    def __init__(self, message):
        self.message = message

        super().__init__(message)

    def __str__(self):
        return self.message



from typing import Optional

import click

from todotree.ConsolePrefixes import ConsolePrefixes


class ConsolePrefixesInit(ConsolePrefixes):
    """
    Provides additional features to ConsolePrefixes.
    These are only used in the Init command.
    """
    def __init__(self, enable_colors: bool,
                 console_good: str, console_warn: str, console_error: str,
                 console_info_color: str, console_warn_color: str, console_error_color: str):
        super().__init__(enable_colors,
                         console_good, console_warn, console_error,
                         console_info_color, console_warn_color, console_error_color)
        self.enable_colors = enable_colors
        self.error_prefix = console_error
        self.warning_prefix = console_warn
        self.info_prefix = console_good
        self.info_color = console_info_color
        self.warn_color = console_warn_color
        self.error_color = console_error_color

    @staticmethod
    def from_console_prefixes(c: ConsolePrefixes):
        """Converts from console prefixes."""
        return ConsolePrefixesInit(
            enable_colors=c.enable_colors,
            console_good=c.info_prefix,
            console_warn=c.warning_prefix,
            console_error=c.error_prefix,
            console_info_color=c.info_color,
            console_warn_color=c.warn_color,
            console_error_color=c.error_color
        )

    def prompt_menu(self, question: str, answers: list, custom_answer: Optional[str] = None) -> str | int:
        """
        Ask the end user for input using a menu.
        @param question: The question to ask the user.
        @param answers: The predefined answers that the user can select.
        @param custom_answer: The custom option Text. Do not append with a dot.
            If the user chooses the Custom answer option. The user will be prompted to fill in the custom answer.
        @return: The answer that the user chose.
        If the user chooses a predefined option, it will be number of that option.
        It will be a string if the answer came from the custom option.
        """
        text = question + "\n"
        for i, answer in enumerate(answers):
            text += f"  [{i}]  {str(answer)}\n"
        # `i` is now the length of the answers list.
        custom_number = len(answers)
        if custom_answer:
            text += f"  [{custom_number}]  {custom_answer}.\n"

        choices = [str(i) for i in range(0, custom_number + 1)]
        answer_int = int(self.prompt(text=text, show_choices=False, type=click.Choice(choices)))
        if custom_answer and answer_int == custom_number:
            answer = self.prompt(text=f"{custom_answer} option chosen. Please type it in.")
            self.confirm(text=f"Is this correct?: {answer}")
            return answer
        else:
            self.confirm(text=f"Is this correct?: {answers[answer_int]}")
            return answer_int

    def prompt(self, *args, **kwargs):
        """Wrapper for click.prompt"""
        self._emit_prefix(self.warn_color, self.warning_prefix)
        return click.prompt(*args, **kwargs)

    def confirm(self, *args, **kwargs):
        """Wrapper for click.confirm"""
        self._emit_prefix(self.warn_color, self.warning_prefix)
        return click.confirm(*args, **kwargs)



from typing import Optional, Callable

from pathlib import Path

import xdg_base_dirs as xdg
from ruamel.yaml import YAML

from todotree.Config.Paths import Paths
from todotree.ConsolePrefixes import ConsolePrefixes
from todotree.Errors.ConfigFileNotFound import ConfigFileNotFound
from todotree.Config.GitHandler import GitHandler
from todotree.Config.TreePrint import TreePrint


class Config:
    """
    The configuration of todotree.

    For a user guide, please refer to the content in `examples/config.yaml` instead.
    Not all values are mapped 1-to-1.
    """

    def __init__(self):
        #  Main variables.
        self.config_file: Optional[Path] = None
        """Path to the config file."""

        # Features - Enables or disables certain features.
        self.enable_project_folder: bool = True
        """A value indicating whether to enable the project folder functionality."""

        #  Localization. #
        self.wishlistName: str = "wishlist"
        self.noProjectString: str = "No Project"
        self.noContextString: str = "No Context"
        self.emptyProjectString: str = "> (A) Todo: add next todo for this."

        # Paths.
        self.paths = Paths()

        #  Status Decorators.
        self.console: ConsolePrefixes = ConsolePrefixes(True, ' * ', ' ! ', '!!!',
                                                        'green', 'yellow', 'red')

        # Tree prints.
        self.tree_print: TreePrint = TreePrint(" ├──", " │  ", " └──", "    ")

        # Git functionality.
        self.git = GitHandler("disabled", 1, self.paths.todo_folder, self.console)

        # Debug
        self.__yaml_object = None
        """Raw config object, for debugging."""

    def __str__(self):
        # Print each attribute in the list.
        s = "Configuration:\n"
        for item in self.__dict__:
            s += f"\t{item} = {self.__dict__[item]} \n"
        return s

    def read(self, path: Optional[Path] = None):
        """
        Loads the configuration from the given file at `path`.

        If empty, reads from default locations.
        """
        if path is not None:
            # print(f"Path is not None: {path}") # FUTURE: Verbose logging
            self.read_from_file(path)
            return
        # xdg compliant directory.
        if Path(xdg.xdg_config_home() / "todotree" / "config.yaml").exists():
            # print(f"Path is XDG Config") # FUTURE: Verbose logging
            self.read_from_file(Path(xdg.xdg_config_home() / "todotree" / "config.yaml"))
            return
        # xdg compliant directory if config file is considered "data".
        if Path(xdg.xdg_data_home() / "todotree" / "config.yaml").exists():
            # print(f"Path is XDG DATA") # FUTURE: Verbose logging
            self.read_from_file(Path(xdg.xdg_data_home() / "todotree" / "config.yaml"))
            return
        # No paths: use the defaults.
        return

    def read_from_file(self, file: Path):
        """Reads and parses yaml content from `file`."""
        self.config_file = file
        try:
            with open(file, 'r') as f:
                self._read_from_yaml(f.read())
        except FileNotFoundError as e:
            raise ConfigFileNotFound(f"Config was not found at {self.config_file}") from e

    def _read_from_yaml(self, yaml_content: str):
        """Reads and overrides config settings defined in `yaml_content`."""
        # Convert yaml to python object.
        self.__yaml_object = YAML().load(yaml_content)
        if self.__yaml_object is None:
            return
        # Map each item to the self config.
        self.__section('main', self.__section_main)
        self.__section('paths', self.__section_paths)
        self.__section('localization', self.__section_localization)
        self.__section('decorators', self.__section_decorators)
        self.__section('tree', self.__section_tree)
        self.__section('git', self.__section_git)

    def __section(self, section_name: str, section_parser: Callable):
        """
        Checks if the section exists and passes it to the sub parser.
        """
        if self.__yaml_object.get(section_name) is not None:
            self.console.verbose(f"[{section_name}] section found, reading it.")
            # Parse the actual values.
            section_parser()
        else:
            self.console.verbose(f"[{section_name}] section not found, skipping it.")

    def __section_tree(self):
        self.tree_print.read_from_dict(self.__yaml_object['tree'])

    def __section_decorators(self):
        self.console.read_from_dict(self.__yaml_object['decorators'])

    def __section_localization(self):
        self.emptyProjectString = self.__yaml_object['localization'].get('empty_project', self.emptyProjectString)
        self.wishlistName = self.__yaml_object['localization'].get('wishlist_name', self.wishlistName)
        self.noProjectString = self.__yaml_object['localization'].get('no_project', self.noProjectString)
        self.noContextString = self.__yaml_object['localization'].get('no_context', self.noContextString)

    def __section_paths(self):
        self.paths.read_from_dict(self.__yaml_object['paths'])
        self.git.todo_folder = self.paths.todo_folder

    def __section_main(self):
        self.enable_project_folder = self.__yaml_object['main'].get('enable_project_folder',
                                                                    self.enable_project_folder)

        # Postprocessing.
        if self.__yaml_object['main'].get("verbose", self.console.is_verbose()):
            self.console.set_verbose()
        if self.__yaml_object['main'].get("quiet", self.console.is_quiet()):
            self.console.set_quiet()

    def __section_git(self):
        self.git.read_from_dict(self.__yaml_object['git'])

    def write_config(self):
        """
        Writes the configuration file to the path specified in `self.config_file`
        """
        # Local import so that the file is not loaded in each time the program is started.
        from importlib import resources as import_resources
        from .. import examples
        # https://stackoverflow.com/questions/6028000/how-to-read-a-static-file-from-inside-a-python-package
        example_config_path = (import_resources.files(examples) / "config.yaml")

        with example_config_path.open("r") as f:
            example_contents = f.read()

        yaml_structure = YAML().load(example_contents)
        # Change features.
        self.tree_print.apply_to_dict(yaml_structure['tree'])
        self.git.apply_to_dict(yaml_structure['git'])
        self.console.apply_to_dict(yaml_structure['decorators'])
        self.paths.apply_to_dict(yaml_structure["paths"])

        # Main.
        yaml_structure['main']['verbose'] = self.console.is_verbose()
        yaml_structure['main']['quiet'] = self.console.is_quiet()
        yaml_structure['main']['enable_project_folder'] = self.enable_project_folder

        # Localization.
        yaml_structure['localization']['wishlist_name'] = self.wishlistName
        yaml_structure['localization']['no_project'] = self.noProjectString
        yaml_structure['localization']['no_context'] = self.noContextString
        yaml_structure['localization']['empty_project'] = self.emptyProjectString

        # Write the file.
        Path(self.config_file).touch()
        write_yaml_settings = YAML()

        write_yaml_settings.dump(yaml_structure, stream=self.config_file)



from pathlib import Path

import xdg_base_dirs

from todotree.Config.Config import Config
from todotree.ConsolePrefixesInit import ConsolePrefixesInit


class Init:
    # Note: Does not inherit AbstractCommand!

    def __init__(self):
        self.config: Config = Config()
        """Configuration."""
        self.action_queue = []
        """List of actions to execute."""
        self.console: ConsolePrefixesInit = ConsolePrefixesInit.from_console_prefixes(self.config.console)

    def run(self):
        self.intro()
        is_git_cloned = False  # self.git_clone() implement.
        if not is_git_cloned:
            self.determine_console_prefixes()
            self.determine_config_location()
            self.determine_folder_location()
            self.enable_git()
            self.enable_project()
            self.example_todo()
            self.generate_config()
        # Write results of the answers to the locations.
        answer = self.console.prompt("Are you sure you want to write the config file?")
        if answer:
            self.config.write_config()

    def intro(self):
        self.console.info("This will configure todotree and generate the necessary files to run it.")
        self.console.info("If at any point you wish to stop, you can do so with Ctrl-C.")
        self.console.info("Also, all options can be changed afterwards in the config.yaml if you change your mind.")

    def determine_config_location(self):
        question = "Where do you want to store config.yaml?"
        answers = [
            Path(xdg_base_dirs.xdg_config_home() / "todotree" / "config.yaml"),
            Path(xdg_base_dirs.xdg_data_home() / "todotree" / "config.yaml")
        ]
        answer = self.console.prompt_menu(question, answers, "Custom location")
        if isinstance(answer, int):
            self.config.config_file = answers[answer]
        else:
            self.config.config_file = answer
            self.console.warning("Note: You need to supply todotree each time with the config location using "
                                 "--config-file")
            self.console.warning("It is advised to alias this in your profile for example in ~/.profile or $PROFILE.")
        self.console.info(f"Set the config location to {self.config.config_file}")

    def determine_console_prefixes(self):
        self.console.enable_colors = self.console.confirm(text="Do you want colors on your console?")
        question = "How do you want the decorations?"
        answers = [
            f"Default      | {self.console.info_prefix} | {self.console.warning_prefix} | {self.console.error_prefix} |"
            ,
            "Gentoo style |  *  |  *  |  *  |"
        ]
        answer = self.console.prompt_menu(question, answers)
        if answer == 1:
            # Then we set gentoo style.
            self.console.info_prefix, self.console.warning_prefix, self.console.error_prefix = ' * '

    def determine_folder_location(self):
        question = "Where do you want to store config.yaml?"
        answers = [
            Path(xdg_base_dirs.xdg_config_home() / "todotree" / "config.yaml"),
            Path(xdg_base_dirs.xdg_data_home() / "todotree" / "config.yaml")
        ]
        answer = self.console.prompt_menu(question, answers, "Custom location")
        if isinstance(answer, int):
            self.config.config_file = answers[answer]
        else:
            self.config.config_file = answer
            self.console.warning("Note: You need to supply todotree each time with the config location using "
                                 "--config-file")
            self.console.warning("It is advised to alias this in your profile i.e. (.profile or $PROFILE)")
        self.console.info(f"Set the config location to {self.config.config_file}")

    def enable_git(self):
        enable_git = self.console.confirm(text="Do you want to enable git?")
        if not enable_git:
            git_mode = "disabled"
        else:
            enable_remote = self.console.confirm(text="Do you want to work with a remote repo as well?")
            git_mode = "remote" if enable_remote else "local"
        self.config.git.git_mode = git_mode

    def example_todo(self):
        if self.console.confirm("Do you want to have your todo file filled with some examples?"):
            self.action_queue += "ExampleTodo"

    def enable_project(self):
        self.console.info("The project_directory functionality adds the projects in a given folder.")
        self.console.info("It will also add tasks if there is no task with the given project name in todo.txt,")
        self.console.info("reminding you that that project is stalled.")
        self.config.enable_project_folder = self.console.confirm(
            "Do you want to enable the project_directory functionality?")

        if self.config.enable_project_folder:
            self.config.project_tree_folder = self.console.prompt("What will the location of your projects be?",
                                                                  default=self.config.project_tree_folder)

    def generate_config(self):
        """Generate the example config with the chosen values."""
        pass

    def git_clone(self):
        answer = self.console.confirm("Do you already have an existing todotree folder on git?")
        if answer:
            self.console.prompt("Please enter the git clone url to clone from.")
            # FUTURE: implement git clone.
        return answer




# NOTE: this file cannot be in a class. See: https://github.com/pallets/click/issues/601
# https://click.palletsprojects.com/en/8.1.x/commands/#nested-handling-and-contexts

from pathlib import Path
from typing import List, Tuple

import click

# Commands are imported when the function is run, to increase performance.
from todotree.Addons import Addons
from todotree.Taskmanager import Taskmanager
from todotree.Config.Config import Config
from todotree.MainUtils import MainUtils

# GLOBALS. #
config = Config()
addons = Addons(config)
taskManager = Taskmanager(config)


@click.group()
@MainUtils.common_options
def root(**kwargs):
    """
    The main list of todotree's command.

    The last option in the command will be in effect, so
       todotree --todo-file task.txt list --todo-file todo.txt
    will read from todo.txt and _not_ from task.txt
    """
    # ^ This text also shows up in the help command.
    # Parse options given before the command.
    # So the `--todo-file task.txt` option in the help text example.
    MainUtils.parse_common_options(config, **kwargs)


@root.command("add", short_help="Add a task to the task list")
@click.argument("task", type=str, nargs=-1)
@MainUtils.common_options
def add(task: Tuple, **kwargs):
    MainUtils.initialize(config, **kwargs)
    from todotree.Commands.Add import Add

    Add(config, taskManager).run(task)


@root.command("addx", short_help="Add a task and immediately mark it as done")
@click.argument("task", type=str, nargs=-1)
@MainUtils.common_options
def add_x(task: Tuple, **kwargs):
    """
    Adds a completed task to done.txt. The task is not added to todo.txt.
    :param task: The task to add to done.txt.
    """
    MainUtils.initialize(config, **kwargs)
    from todotree.Commands.AddX import AddX

    AddX(config, taskManager).run(task)


@root.command("append", short_help="append `append_string` to `task_nr`")
@click.argument("task_nr", type=int)
@click.argument("append_string", type=str, nargs=-1)
@MainUtils.common_options
def append(task_nr: int, append_string: Tuple[str], **kwargs):
    """
    Appends the contents of append_string to the task represented by task_nr.
    A space is inserted between the two tasks, so you do not have to worry that words aretogether.

    Example: todotree append 1 "some text"
    """
    MainUtils.initialize(config, **kwargs)
    from todotree.Commands.Append import Append

    Append(config, taskManager).run(task_nr, append_string)


@root.command("cd", short_help="print directory of the todo.txt directory")
@MainUtils.common_options
def cd(**kwargs):
    """print directory of the todo.txt directory"""
    MainUtils.initialize(config, **kwargs)
    from todotree.Commands.Cd import Cd

    Cd(config, taskManager).run()


@root.command("context", short_help="list task in a tree, by context")
@MainUtils.common_options
def context(**kwargs):
    """list a tree, of which the first node is the context, the second nodes are the tasks"""
    MainUtils.initialize(config, **kwargs)
    from todotree.Commands.Context import Context

    Context(config, taskManager).run()


@root.command("do", short_help="mark task as done and move it to the done.txt")
@click.argument("task_numbers", type=list, nargs=-1)  # type=list[int]
@MainUtils.common_options
def do(task_numbers: List[Tuple], **kwargs):
    """
    Mark tasks as done, therefor moving them to done.txt with a date stamp of today.
    :param task_numbers: The list of tasks which are completed.
    """
    MainUtils.initialize(config, **kwargs)
    from todotree.Commands.Do import Do

    Do(config, taskManager).run(task_numbers)


@root.command("due", short_help="List tasks by their due date")
@MainUtils.common_options
def due(**kwargs):
    """List tasks in a tree by their due date."""
    MainUtils.initialize(config, **kwargs)
    from todotree.Commands.Due import Due

    Due(config, taskManager).run()


@root.command("edit", short_help="open the todo.txt in an editor.")
@MainUtils.common_options
def edit(**kwargs):
    """
    Open the todo.txt in an editor for manual editing of tasks.
    This is useful when you need to modify a lot of tasks, which would be complicated when doing it with todotree.
    """
    MainUtils.initialize(config, **kwargs)
    from todotree.Commands.Edit import Edit

    Edit(config, taskManager).run()


@root.command("filter", short_help="only show tasks containing the search term.")
@click.argument("search_term")
@MainUtils.common_options
def filter_list(search_term, **kwargs):
    """
    Only show tasks which have search term in them. This can also be a keyword.

    :param search_term: The term to search.
    """
    MainUtils.initialize(config, **kwargs)
    from todotree.Commands.Filter import Filter

    Filter(config, taskManager).run(search_term)


@root.command("help", short_help="Print help text.")
@click.pass_context
def help_text(ctx: click.Context):
    print(ctx.parent.get_help())


@root.command("init", short_help="Initialize folder for first use.")
@MainUtils.common_options
def init(**kwargs):
    """
    Initializes the files and folders needed for a functioning todotree according to the user's prompts.
    """
    MainUtils.initialize(config, **kwargs)
    from todotree.Commands.Init import Init

    Init().run()


@root.command("list", short_help="List tasks")
@MainUtils.common_options
def list_tasks(**kwargs):
    """
    Print a flat list of tasks, sorted by their priority.
    """
    MainUtils.initialize(config, **kwargs)
    from todotree.Commands.List import List

    MainUtils.parse_common_options(config, **kwargs)
    List(config, taskManager).run()


@root.command("list_done", short_help="List tasks which are marked as done")
@MainUtils.common_options
def list_done(**kwargs):
    """
    List tasks which are marked as done. The numbers can be used with the revive command.
    """
    MainUtils.initialize(config, **kwargs)
    from todotree.Commands.ListDone import ListDone

    ListDone(config, taskManager).run()


@root.command("print_raw", short_help="print todo.txt without any formatting or filtering")
@MainUtils.common_options
def print_raw(**kwargs):
    """
    Output the todo.txt without any processing.
    This is equivalent to `cat $(todo cd)/todo.txt` in bash.
    """
    MainUtils.initialize(config, **kwargs)
    from todotree.Commands.PrintRaw import PrintRaw

    PrintRaw(config, taskManager).run()


@root.command("priority", short_help="set new priority to task")
@click.argument("task_number", type=int)
@click.argument("new_priority", type=str)
@MainUtils.common_options
def priority(task_number, new_priority, **kwargs):
    """
    Adds or updates the priority of the task.
    :param task_number: The task to re-prioritize.
    :param new_priority: The new priority.
    """
    MainUtils.initialize(config, **kwargs)
    from todotree.Commands.Priority import Priority

    Priority(config, taskManager).run(task_number, new_priority)


@root.command("project", short_help="print tree by project")
@MainUtils.common_options
def project(**kwargs):
    """
    Print the task list in a tree by project.
    """
    MainUtils.initialize(config, **kwargs)
    from todotree.Commands.Project import Project

    Project(config, taskManager).run()


@root.command("revive", short_help="Revive a task that was accidentally marked as done.")
@click.argument("done_number", type=int)
@MainUtils.common_options
def revive(done_number, **kwargs):
    """
    Move a task from done.txt to todo.txt.
    The `done_number` can be found using the `list_done` command.
    :param done_number: The number of the task to revive.
    """
    MainUtils.initialize(config, **kwargs)
    from todotree.Commands.Revive import Revive

    Revive(config, taskManager).run(done_number)


@root.command("schedule", short_help="hide task until date.")
@click.argument("task_number", type=int)
@click.argument("new_date", type=str, nargs=-1)
@MainUtils.common_options
def schedule(task_number: int, new_date: Tuple[str], **kwargs):
    """
    hide the task until the date given. If new_date is not in ISO format (yyyy-mm-dd) such as "Next Wednesday",
    then it tries to figure out the date with the `date` program, which is only in linux.
    """
    MainUtils.initialize(config, **kwargs)
    from todotree.Commands.Schedule import Schedule

    Schedule(config, taskManager).run(new_date, task_number)


@root.command("version")
@MainUtils.common_options
def version(**kwargs):
    MainUtils.initialize(config, **kwargs)
    from todotree.Commands.Version import Version

    Version(config, taskManager).run()


@root.command("addons", short_help="Run an addon script")
@click.argument("command", type=str)
@MainUtils.common_options
def addons_command(command: str, **kwargs):
    """
    Run an addon script.
    The script must be in the addons_folder. It can be any language you like: It does not have to be python.
    However, it must have the executable bit set.

    :param command: The script/command to run.
    """
    MainUtils.initialize(config, **kwargs)
    result = Addons(config).run(command)
    click.echo(result.stdout, nl=False)
    if result.returncode != 0:
        click.echo(result.stderr)
    config.git.commit_and_push(f"addons {command}")


#  End Region Command Definitions.
#  Setup Click

CONTEXT_SETTINGS: dict = dict(help_option_names=["-h", "--help"])
"""Click context settings. See https://click.palletsprojects.com/en/8.1.x/complex/ for more information."""
cli: click.CommandCollection = click.CommandCollection(sources=[root], context_settings=CONTEXT_SETTINGS)
"""Command Collection defining defaults. https://click.palletsprojects.com/en/8.1.x/api/#click.CommandCollection ."""

if __name__ == "__main__":
    cli()



from todotree.Config.Config import Config


class DoneFileNotFound(BaseException):
    """
    Represents an error indicating that the done.txt file is not found.
    """

    def __init__(self, message):
        self.message = message

        super().__init__(message)

    def __str__(self):
        return self.message

    def echo_and_exit(self, config: Config):
        config.console.error("The done.txt could not be found.")
        config.console.error(f"It searched at the following location: {config.paths.done_file}")
        config.console.verbose(self)
        exit(1)



from typing import Tuple, List

import click

from todotree.Commands.AbstractCommand import AbstractCommand
from todotree.Errors.DoneFileNotFound import DoneFileNotFound
from todotree.Errors.TodoFileNotFound import TodoFileNotFound


class Do(AbstractCommand):

    def run(self, task_numbers: List[Tuple]):
        # Convert to ints. Task numbers is a list of tuples. Each tuple contains one digit of the number.
        new_numbers: List[int] = []
        for task_tuple in task_numbers:
            new_number: str = ""
            for task_digit in task_tuple:
                new_number += task_digit
            new_numbers.append(int(new_number))
        # Write back to old value.
        task_numbers = new_numbers
        # Marking something as Done cannot be done with fancy imports
        # So we disable them.
        self.config.enable_project_folder = False
        try:
            self.taskManager.import_tasks()
        except TodoFileNotFound as e:
            e.echo_and_exit(self.config)
        try:
            completed_tasks = self.taskManager.mark_as_done(task_numbers)
        except DoneFileNotFound as e:
            e.echo_and_exit(self.config)
            exit(1)  # For IDE hinting.
        # Print the results
        self.config.console.info("Tasks marked as done:")
        for task in completed_tasks:
            click.echo(task)
        self.config.git.commit_and_push("do")



import click

from todotree.Commands.AbstractCommand import AbstractCommand
from todotree.Errors.DoneFileNotFound import DoneFileNotFound
from todotree.Errors.TodoFileNotFound import TodoFileNotFound


class Revive(AbstractCommand):
    def run(self, done_number: int):
        try:
            click.echo(self.taskManager.revive_task(done_number))
        except TodoFileNotFound as e:
            e.echo_and_exit(self.config)
        except DoneFileNotFound as e:
            e.echo_and_exit(self.config)
        self.config.git.commit_and_push("revive")



import datetime
import math
import os
import pathlib
import shutil
from tempfile import (  # https://docs.python.org/3/library/tempfile.html
    NamedTemporaryFile,
)
from typing import Callable

from todotree.Errors.TodoFileNotFound import TodoFileNotFound
from todotree.Errors.DoneFileNotFound import DoneFileNotFound
from todotree.Errors.ProjectFolderError import ProjectFolderError
from todotree.Task.ConsoleTask import ConsoleTask
from todotree.Task.Task import Task
from todotree.Tree import Tree
from todotree.Config.Config import Config
from todotree.Task.DoneTask import DoneTask


class Taskmanager:
    """Manages the tasks."""

    @property
    def number_of_digits(self) -> int:
        """Property which defines the number of digits of the task in the task list with the highest number."""
        return int(math.ceil(math.log(len(self.task_list) + 1, 10)))

    def __init__(self, configuration: Config = Config()):
        """Initializes a task manager, which manages the tasks."""

        self.task_list: list[Task] = []
        """List Without empty lines"""

        self.config: Config = configuration
        """Configuration settings for the application."""

    # Imports

    def import_tasks(self):
        """
        Import the tasks and projects from files.
        """
        self._import_from_todo_txt()
        if self.config.enable_project_folder:
            self._import_project_folder()

    def _import_from_todo_txt(self):
        """
        Imports the todos from the file to the Taskmanager object.
        """
        try:
            with self.config.paths.todo_file.open('r') as f:
                content = f.readlines()
                self.task_list = []
                for i, task in enumerate(content):
                    # Skip empty lines.
                    if task.strip() == "":
                        continue
                    self.task_list.append(Task(i + 1, task.strip()))
        except FileNotFoundError as e:
            raise TodoFileNotFound("Todo not found.") from e

    def filter_t_date(self):
        """
        Removes Tasks from the Taskmanager where the t:date is later than today.
        """
        for task in self.task_list.copy():
            if task.t_date is None:
                # It does not have a t date.
                continue
            if task.t_date > datetime.date.today():
                self.task_list.remove(task)

    def filter_block(self):
        """
        Removes Tasks from the Taskmanager where the task is blocked by some other task.
        """
        # Detect all block items.
        block_list = []
        for task in self.task_list:
            block_list.append(task.blocks)
        # Flatten the list of lists to a normal list.
        block_list_items = [item for sublist in block_list for item in sublist]
        # Blocked / Blocked by filtering
        for task in self.task_list.copy():
            # Check if there is a task which is blocked by another task.
            if list(filter(lambda x: x in block_list_items, task.blocked)):
                self.task_list.remove(task)

    def _import_project_folder(self):
        """
        Adds the directory to the project tree.
        """
        try:
            for i in pathlib.Path(self.config.paths.project_tree_folder).glob("*/"):
                proj = os.path.basename(i)
                # Check if there is a task with that project.
                does_project_task_exist = any(proj in ttt.projects for ttt in self.task_list)
                if not does_project_task_exist:
                    # add gibberish task as placeholder.
                    tmp_task = Task(-1, self.config.emptyProjectString)
                    tmp_task.projects = [proj]
                    self.task_list.append(tmp_task)
        except FileNotFoundError as e:
            raise ProjectFolderError(f"An error occurred while processing the projects folder: {e}")

    def list_done(self):
        """
        List the done.txt with numbers to revive them.
        """
        try:
            with self.config.paths.done_file.open(mode="r") as f:
                lines = f.readlines()
        except FileNotFoundError as e:
            raise DoneFileNotFound from e
        for i, line in reversed(list(enumerate(lines))):
            print(i, line, end="")

    def filter_by_string(self, filter_string):
        """
        Filters the task list by whether `filter_string` occurred in the task.
        """
        self.task_list = [item for item in self.task_list if filter_string in item.task_string]

    def print_by_due(self):
        """
        Prints the tree of due dates.
        :return: A string containing the due tree.
        """
        self.filter_t_date()
        self.filter_block()
        # Convert to ConsoleTask.
        tasks = [ConsoleTask(task.i, task.task_string, total_digits=self.number_of_digits) for task in self.task_list]
        # Print function.
        print_function: Callable[[str, ConsoleTask], str] = lambda _, task: task.print_due()
        return str((Tree(tasks, "due_date_band", treeprint=self.config.tree_print, print_func=print_function,
                         root_name="Due Dates")))

    def print_project_tree(self) -> str:
        """
        print the project tree.
        :return: A string containing the project tree.
        """
        self.filter_t_date()
        self.filter_block()
        # Convert to ConsoleTask.
        tasks = [ConsoleTask(task.i, task.task_string, total_digits=self.number_of_digits) for task in self.task_list]
        # Print function.
        print_function: Callable[[str, ConsoleTask], str] = \
            lambda project_to_exclude, task: task.print_project(project_to_exclude)
        return str((Tree(tasks, "projects", treeprint=self.config.tree_print, print_func=print_function,
                         root_name="Projects", empty_string=self.config.noProjectString)))

    def print_context_tree(self) -> str:
        """
        print the context tree.
        :return: A string containing the context tree.
        """
        self.filter_t_date()
        self.filter_block()
        tasks = [ConsoleTask(task.i, task.task_string, total_digits=self.number_of_digits) for task in self.task_list]
        # Print function.
        print_function: Callable[[str, ConsoleTask], str] = \
            lambda context_to_exclude, task: task.print_context(context_to_exclude)
        return str((Tree(tasks, "contexts", treeprint=self.config.tree_print, print_func=print_function,
                         root_name="Contexts", empty_string=self.config.noContextString)))

    #  Write Only Methods

    def write_todo_to_file(self):
        """
        Write the results back to a file.
        """
        # Write to new file.
        self.task_list.sort(key=lambda x: x.i)
        #  Delete=false is needed for windows, I hope that somebodies temp folder won't be clobbered with this...
        try:
            with NamedTemporaryFile("w+t", newline="", delete=False) as temp_file:
                # may strip new lines by using task list.
                for n in self.task_list:
                    temp_file.write(str(n.task_string))
                    temp_file.write("\n")
                temp_file.flush()
                shutil.copy(temp_file.name, self.config.paths.todo_file)
        except FileNotFoundError as e:
            raise TodoFileNotFound from e

    def add_or_update_t_date(self, date: str, task_number: int):
        """
        Adds or changes the t:date
        :param date: The new t:date to add.
        :param task_number: The task number to add the new date to.
        :return: The updated task.
        """
        self.task_list[task_number - 1].add_or_update_t_date(date)
        self.write_todo_to_file()
        return self.task_list[task_number - 1]

    def add_or_update_priority(self, priority: str, task_number):
        """
        Adds or changes the priority of a task.
        :param priority: The new priority.
        :param task_number: The number of the task.
        :return: The new task.
        """
        self.task_list[task_number - 1].add_or_update_priority(priority)
        self.write_todo_to_file()
        return self.task_list[task_number - 1]

    def add_task_to_file(self, tsk: str) -> int:
        """
        Adds a task to a file. Returns the task with its assigned number.
        :param tsk: The task to add.
        :return: The number of the new task.
        """
        try:
            with self.config.paths.todo_file.open(mode="a") as f:
                f.write(tsk)
        except FileNotFoundError as e:
            raise TodoFileNotFound from e
        # return the line number
        return len(self.task_list) + 1

    def append_to_task(self, task_number: int, task_string: str) -> str:
        """
        Appends `task_string` to the task defined by `task_number`.
        :param task_string: The string to append
        :param task_number: The number of the existing task.
        :return: The new merged task.
        """
        self.task_list[task_number - 1] = self.task_list[task_number - 1] + task_string
        self.write_todo_to_file()
        return self.task_list[task_number - 1].task_string

    def mark_as_done(self, line_numbers):
        """
        Marks the task as done, removes it from the task file and adds it to done.txt
        :param line_numbers: The task numbers to mark as done.
        :return: The tasks marked as done.
        """
        # Get and remove completed task
        completed_tasks = []
        line_numbers.sort()
        # Delete the tasks from the end to the start.
        line_numbers.reverse()
        for num in line_numbers:
            completed_tasks.append(self.task_list[num - 1])
            del self.task_list[num - 1]
        # Save the new task list.
        self.write_todo_to_file()
        # Convert completed tasks to done tasks.
        done_tasks = list(map(lambda x: DoneTask.task_to_done(x), completed_tasks))
        # Append the results to done.txt
        try:
            with self.config.paths.done_file.open(mode="a") as f:
                for completed_task in done_tasks:
                    f.write(completed_task + "\n")
        except FileNotFoundError as e:
            raise DoneFileNotFound from e
        return done_tasks

    def revive_task(self, line_number):
        """
        Gets task from done.txt and adds it to the task file.
        The task does not get removed.
        :param line_number: The done number to revive.
        :return: The revived task.
        """
        # Fetch Task
        try:
            with self.config.paths.done_file.open() as f:
                done_tasks = f.readlines()
                print(done_tasks, line_number)
        except FileNotFoundError as e:
            raise DoneFileNotFound from e
        # Add task.
        new_task = DoneTask.task_to_undone(done_tasks[line_number - 1])
        new_task_number = self.add_task_to_file(new_task)
        return ConsoleTask(new_task_number, new_task)

    def __str__(self):
        # Prepare the lists.
        self.filter_t_date()
        self.filter_block()
        self.task_list.sort(key=lambda x: x.priority)
        # Output the list.
        s = ""
        for tsk in [ConsoleTask(task.i, task.task_string, self.number_of_digits) for task in self.task_list]:
            s += str(tsk) + "\n"
        return s



import click

from todotree.Commands.AbstractCommand import AbstractCommand
from todotree.Errors.TodoFileNotFound import TodoFileNotFound


class PrintRaw(AbstractCommand):

    def run(self):
        try:
            with open(self.taskManager.config.paths.todo_file, "r") as f:
                click.echo(f.read())
        except FileNotFoundError:
            TodoFileNotFound("").echo_and_exit(self.config)




from todotree.ConsolePrefixes import ConsolePrefixes


class GitError(Exception):
    """
    Represents an error indicating that something is wrong with the git repo.
    """
    def __init__(self, message):
        self.message = message

        super().__init__(message)

    def __str__(self):
        return self.message

    def warn_and_continue(self, console: ConsolePrefixes):
        console.warning("An error occurred while trying to git pull.")
        console.warning("If this was unexpected, try again with the --verbose flag.")





from todotree.Commands.AbstractCommand import AbstractCommand
from todotree.Errors.DoneFileNotFound import DoneFileNotFound


class ListDone(AbstractCommand):
    def run(self):
        try:
            self.taskManager.list_done()
        except DoneFileNotFound as e:
            e.echo_and_exit(self.config)



from datetime import date

from todotree.Task.Task import Task


class DoneTask:

    @staticmethod
    def task_to_done(tsk):
        """
        Format the task String to a done state, by adding x 2020-02-02 in front of it.
        :param tsk: A task_string.
        :return: A done_string with a today's time stamp.
        """
        if isinstance(tsk, Task):  # Flatten to string
            tsk = tsk.task_string
        done = "x " + str(date.today()) + " " + tsk
        return done

    @staticmethod
    def task_to_undone(tsk):
        """
        Removes the x 2020-02-02 part so the task can be added to the task file list
        :param tsk: A done string.
        :return: A task string.
        """
        # 13 is the number of chars of "x 2020-02-02 "
        return str.strip(tsk[13:])
